/*
 *
 * C2VectParallel implementation
 *
 * Copyright 2020 Torge Matthies
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <stdarg.h>

#include "windef.h"
#include "wine/debug.h"
#include "wine/asm.h"

WINE_DEFAULT_DEBUG_CHANNEL(vcomp);

extern int CDECL omp_get_num_threads(void);
extern int CDECL omp_get_thread_num(void);
extern void CDECL _vcomp_set_num_threads(int num_threads);
extern void WINAPIV _vcomp_fork(BOOL ifval, int nargs, void *wrapper, ...);

#if defined(__i386__)

extern void CDECL C2VectParallel_call_func(void (CDECL * func)(int, int, ...), int nargs,
										   int arg1, int arg2, __ms_va_list args);
__ASM_GLOBAL_FUNC( C2VectParallel_call_func,
                   "pushl %ebp\n\t"
                   __ASM_CFI(".cfi_adjust_cfa_offset 4\n\t")
                   __ASM_CFI(".cfi_rel_offset %ebp,0\n\t")
                   "movl %esp,%ebp\n\t"
                   __ASM_CFI(".cfi_def_cfa_register %ebp\n\t")
                   "pushl %esi\n\t"
                   __ASM_CFI(".cfi_rel_offset %esi,-4\n\t")
                   "pushl %edi\n\t"
                   __ASM_CFI(".cfi_rel_offset %edi,-8\n\t")
                   "movl 12(%ebp),%edx\n\t"
                   "movl %esp,%edi\n\t"
                   "shll $2,%edx\n\t"
                   "jz 1f\n\t"
                   "subl %edx,%edi\n\t"
                   "andl $~15,%edi\n\t"
                   "movl %edi,%esp\n\t"
                   "movl 16(%ebp),%ecx\n\t"
                   "movl %ecx,0(%edi)\n\t"
                   "movl 20(%ebp),%ecx\n\t"
                   "movl %ecx,4(%edi)\n\t"
                   "addl $8,%edi\n\t"
                   "movl 12(%ebp),%ecx\n\t"
                   "subl $2,%ecx\n\t"
                   "movl 24(%ebp),%esi\n\t"
                   "cld\n\t"
                   "rep; movsl\n"
                   "1:\tcall *8(%ebp)\n\t"
                   "leal -8(%ebp),%esp\n\t"
                   "popl %edi\n\t"
                   __ASM_CFI(".cfi_same_value %edi\n\t")
                   "popl %esi\n\t"
                   __ASM_CFI(".cfi_same_value %esi\n\t")
                   "popl %ebp\n\t"
                   __ASM_CFI(".cfi_def_cfa %esp,4\n\t")
                   __ASM_CFI(".cfi_same_value %ebp\n\t")
                   "ret" )

#elif defined(__x86_64__)

extern void CDECL C2VectParallel_call_func(void (CDECL * func)(int, int, ...), int nargs,
										   int arg1, int arg2, __ms_va_list args);
__ASM_GLOBAL_FUNC( C2VectParallel_call_func,
                   "pushq %rbp\n\t"
                   __ASM_SEH(".seh_pushreg %rbp\n\t")
                   __ASM_CFI(".cfi_adjust_cfa_offset 8\n\t")
                   __ASM_CFI(".cfi_rel_offset %rbp,0\n\t")
                   "movq %rsp,%rbp\n\t"
                   __ASM_SEH(".seh_setframe %rbp,0\n\t")
                   __ASM_CFI(".cfi_def_cfa_register %rbp\n\t")
                   "pushq %rsi\n\t"
                   __ASM_SEH(".seh_pushreg %rsi\n\t")
                   __ASM_CFI(".cfi_rel_offset %rsi,-8\n\t")
                   "pushq %rdi\n\t"
                   __ASM_SEH(".seh_pushreg %rdi\n\t")
                   __ASM_SEH(".seh_endprologue\n\t")
                   __ASM_CFI(".cfi_rel_offset %rdi,-16\n\t")
                   "movq %rcx,%rax\n\t"
                   "movq $4,%rcx\n\t"
                   "cmp %rcx,%rdx\n\t"
                   "cmovgq %rdx,%rcx\n\t"
                   "leaq 0(,%rcx,8),%rdx\n\t"
                   "subq %rdx,%rsp\n\t"
                   "andq $~15,%rsp\n\t"
                   "subq $2,%rcx\n\t"
                   "movq %rsp,%rdi\n\t"
                   "movq %r8,0(%rdi)\n\t"
                   "movq %r9,8(%rdi)\n\t"
                   "addq $16,%rdi\n\t"
                   "movq 48(%rbp),%rsi\n\t"
                   "rep; movsq\n\t"
                   "movq 0(%rsp),%rcx\n\t"
                   "movq 8(%rsp),%rdx\n\t"
                   "movq 16(%rsp),%r8\n\t"
                   "movq 24(%rsp),%r9\n\t"
                   "callq *%rax\n\t"
                   "leaq -16(%rbp),%rsp\n\t"
                   "popq %rdi\n\t"
                   __ASM_CFI(".cfi_same_value %rdi\n\t")
                   "popq %rsi\n\t"
                   __ASM_CFI(".cfi_same_value %rsi\n\t")
                   __ASM_CFI(".cfi_def_cfa_register %rsp\n\t")
                   "popq %rbp\n\t"
                   __ASM_CFI(".cfi_adjust_cfa_offset -8\n\t")
                   __ASM_CFI(".cfi_same_value %rbp\n\t")
                   "ret")

/*#elif defined(__arm__)

extern void CDECL C2VectParallel_call_func(void (CDECL * func)(int, int, ...), int nargs,
										   int arg1, int arg2, __ms_va_list args);
__ASM_GLOBAL_FUNC( C2VectParallel_call_func,
                   ".arm\n\t"
                   "push {r4, r5, LR}\n\t"
                   "mov r4, r0\n\t"
                   "mov r5, SP\n\t"
                   "lsl r3, r1, #2\n\t"
                   "cmp r3, #0\n\t"
                   "beq 5f\n\t"
                   "sub SP, SP, r3\n\t"
                   "tst r1, #1\n\t"
                   "subeq SP, SP, #4\n\t"
                   "1:\tsub r3, r3, #4\n\t"
                   "ldr r0, [r2, r3]\n\t"
                   "str r0, [SP, r3]\n\t"
                   "cmp r3, #0\n\t"
                   "bgt 1b\n\t"
                   "cmp r1, #1\n\t"
                   "bgt 2f\n\t"
                   "pop {r0}\n\t"
                   "b 5f\n\t"
                   "2:\tcmp r1, #2\n\t"
                   "bgt 3f\n\t"
                   "pop {r0-r1}\n\t"
                   "b 5f\n\t"
                   "3:\tcmp r1, #3\n\t"
                   "bgt 4f\n\t"
                   "pop {r0-r2}\n\t"
                   "b 5f\n\t"
                   "4:\tpop {r0-r3}\n\t"
                   "5:\tblx r4\n\t"
                   "mov SP, r5\n\t"
                   "pop {r4, r5, PC}" )

#elif defined(__aarch64__)

extern void CDECL C2VectParallel_call_func(void (CDECL * func)(int, int, ...), int nargs,
										   int arg1, int arg2, __ms_va_list args);
__ASM_GLOBAL_FUNC( C2VectParallel_call_func,
                   "stp x29, x30, [SP,#-16]!\n\t"
                   "mov x29, SP\n\t"
                   "mov x9, x0\n\t"
                   "cbz w1, 4f\n\t"
                   "lsl w8, w1, #3\n\t"
                   "cmp w8, #64\n\t"
                   "b.ge 1f\n\t"
                   "mov w8, #64\n"
                   "1:\ttbz w8, #3, 2f\n\t"
                   "add w8, w8, #8\n"
                   "2:\tsub x10, x29, x8\n\t"
                   "mov sp, x10\n"
                   "3:\tldr x0, [x2], #8\n\t"
                   "str x0, [x10], #8\n\t"
                   "subs w1, w1, #1\n\t"
                   "b.ne 3b\n\t"
                   "ldp x0, x1, [sp], #16\n\t"
                   "ldp x2, x3, [sp], #16\n\t"
                   "ldp x4, x5, [sp], #16\n\t"
                   "ldp x6, x7, [sp], #16\n"
                   "4:\tblr x9\n\t"
                   "mov SP, x29\n\t"
                   "ldp x29, x30, [SP], #16\n\t"
                   "ret\n" )*/

#else

static void CDECL C2VectParallel_call_func(void (CDECL * func)(int, int, ...), int nargs,
										   int arg1, int arg2, __ms_va_list args)
{
    ERR("Not implemented for this architecture\n");
}

#endif

struct C2VectParallel_data {
    int start;
    int end;
    int stride;
    BOOL inclusive;

    void (* func)(int, int, ...);
    int argcnt;
	__ms_va_list valist;
};

void WINAPIV C2VectParallel_callback(struct C2VectParallel_data* data)
{
    int start, end, step, stride, num_threads;

    /* Load the stride into a local variable. */
    stride = data->stride;
    /* Get the number of threads working on the input. */
    num_threads = omp_get_num_threads();

    /* Calculate how much of the input each thread works on. */
    step = data->end - data->start;
    if (step < 0) step = -step;
    if (data->inclusive) step = step + 1;
    step = step + (num_threads - 1);
    step = step / num_threads;

    /* Calculate the total number of input elements. */
    step = data->end - data->start;
    if (step < 0) step = -step;
    if (data->inclusive) step = step + 1;
    /* Divide by the number of threads, while rounding up. */
    step = (step + (num_threads - 1)) / num_threads;
    /* Make step a multiple of stride. */
    if (step % stride != 0)
        step += (stride < 0 ? -stride : stride) - step % stride;
    /* Make step negative if stride is negative. */
    if (stride < 0)
        step = -step;

    /* Calculate the start and the end of the range of input this thread
       should be working on. */
    start = data->start + step * omp_get_thread_num();
    end = start + step;

    /* Remove the last element from the range if inclusive is true.
       Needed because otherwise the callback would process the first element
       of the following range too. */
    if (data->inclusive)
        end -= step < 0 ? -1 : 1;

    /* Clamp start and end to the input data range, and just return if
       they are both outside the range. */
    if (step < 0)
    {
        if (end < data->end || (!data->inclusive && end == data->end))
        {
            if (start < data->end || (!data->inclusive && start == data->end))
                return;
            end = data->end;
        }
    }
    else
    {
        if (end > data->end || (!data->inclusive && end == data->end))
        {
            if (start > data->end || (!data->inclusive && start == data->end))
                return;
            end = data->end;
        }
    }

    /* Call the user-provided callback function. */
    C2VectParallel_call_func(data->func, data->argcnt, start, end, data->valist);
}

void WINAPIV C2VectParallel(int start, int end, int stride, BOOL inclusive, unsigned int numChunks,
                            int schedule, void (CDECL * func)(int, int, ...), int argcnt, ...)
{
    struct C2VectParallel_data data;

    TRACE("(%d, %d, %d, %d, %ud, %d, %p, %d, ...)\n", start, end, stride, inclusive, numChunks, schedule, func, argcnt);

    data.start = start;
    data.end = end;
    data.stride = stride;
    data.inclusive = inclusive;
    data.func = func;
    data.argcnt = argcnt;
    __ms_va_start(data.valist, argcnt);

    _vcomp_set_num_threads(numChunks);
    _vcomp_fork(TRUE, 1, C2VectParallel_callback, &data);

    __ms_va_end(data.valist);
}
