/*
 * Copyright 2020 Rémi Bernon for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#ifndef WMPHOTO_PRIVATE_H
#define WMPHOTO_PRIVATE_H

#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "wincodecsdk.h"

typedef long JXR_ERR;

#define WMP_errSuccess 0
#define WMP_errFail -1
#define WMP_errFileIO -102

struct jxrlib_stream
{
    JXR_ERR (CDECL *Read)(struct jxrlib_stream *iface, void *buf, size_t len);
    JXR_ERR (CDECL *Write)(struct jxrlib_stream *iface, const void *buf, size_t len);
    JXR_ERR (CDECL *SetPos)(struct jxrlib_stream *iface, size_t pos);
    JXR_ERR (CDECL *GetPos)(struct jxrlib_stream *iface, size_t *pos);
};

struct jxrlib_decoder
{
    JXR_ERR (CDECL *Initialize)(struct jxrlib_decoder *iface, struct jxrlib_stream *stream);
    JXR_ERR (CDECL *GetPixelFormat)(struct jxrlib_decoder *iface, GUID *guid);
    JXR_ERR (CDECL *GetSize)(struct jxrlib_decoder *iface, INT32 *width, INT32 *height);
    JXR_ERR (CDECL *GetResolution)(struct jxrlib_decoder *iface, FLOAT *resx, FLOAT *resy);
    JXR_ERR (CDECL *Copy)(struct jxrlib_decoder *iface, const WICRect *rect, UINT8 *buffer, UINT32 stride);
    JXR_ERR (CDECL *GetFrameCount)(struct jxrlib_decoder *iface, UINT32 *count);
    JXR_ERR (CDECL *SelectFrame)(struct jxrlib_decoder *iface, UINT32 index);
    JXR_ERR (CDECL *Release)(struct jxrlib_decoder **iface);
};

struct jxrlib_funcs
{
    HRESULT (CDECL *decoder_create)(struct jxrlib_decoder **out);
};

#endif /* WMPHOTO_PRIVATE_H */
